@@ -1,4 +1,4 @@
-;;; diary-lib.el --- diary functions.
+;;; diary-lib.el --- diary functions
 
 ;; Copyright (C) 1989, 1990, 1992, 1993, 1994, 1995 Free Software
 ;; Foundation, Inc.
@@ -197,6 +197,7 @@ It is the standard syntax table used in Fundamental mode, but with the
 syntax of `*' changed to be a word constituent.")
 
 (modify-syntax-entry ?* "w" diary-syntax-table)
+(modify-syntax-entry ?: "w" diary-syntax-table)
 
 (defun list-diary-entries (date number)
   "Create and display a buffer containing the relevant lines in diary-file.
@@ -254,10 +255,12 @@ These hooks have the following distinct roles:
             (let ((buffer-read-only nil)
                   (diary-modified (buffer-modified-p))
                   (mark (regexp-quote diary-nonmarking-symbol)))
+              ;; First and last characters must be ^M or \n for
+              ;; selective display to work properly
               (goto-char (1- (point-max)))
               (if (not (looking-at "\^M\\|\n"))
                   (progn
-                    (forward-char 1)
+                    (goto-char (point-max))
                     (insert-string "\^M")))
               (goto-char (point-min))
               (if (not (looking-at "\^M\\|\n"))
@@ -290,7 +293,7 @@ These hooks have the following distinct roles:
                            (concat
                             "\\*\\|0*" (int-to-string year)
                             (if abbreviated-calendar-year
-                                (concat "\\|" (int-to-string (% year 100)))
+                                (concat "\\|" (format "%02d" (% year 100)))
                               "")))
                           (regexp
                            (concat
@@ -328,7 +331,7 @@ These hooks have the following distinct roles:
                      (setq d (cdr d)))
                    (or entry-found
                        (not diary-list-include-blanks)
-                       (setq diary-entries-list 
+                       (setq diary-entries-list
                              (append diary-entries-list
                                      (list (list date "" "")))))
                    (setq date
@@ -362,20 +365,29 @@ changing the variable `diary-include-string'."
            (regexp-quote diary-include-string)
            " \"\\([^\"]*\\)\"")
           nil t)
-    (let ((diary-file (substitute-in-file-name
-                       (buffer-substring-no-properties
-                        (match-beginning 2) (match-end 2))))
-          (diary-list-include-blanks nil)
-          (list-diary-entries-hook 'include-other-diary-files)
-          (diary-display-hook 'ignore)
-          (diary-hook nil))
+    (let* ((diary-file (substitute-in-file-name
+                        (buffer-substring-no-properties
+                         (match-beginning 2) (match-end 2))))
+           (diary-list-include-blanks nil)
+           (list-diary-entries-hook 'include-other-diary-files)
+           (diary-display-hook 'ignore)
+           (diary-hook nil)
+           (d-buffer (find-buffer-visiting diary-file))
+           (diary-modified (if d-buffer
+                               (save-excursion
+                                 (set-buffer d-buffer)
+                                 (buffer-modified-p)))))
       (if (file-exists-p diary-file)
           (if (file-readable-p diary-file)
               (unwind-protect
                   (setq diary-entries-list
                         (append diary-entries-list
                                 (list-diary-entries original-date number)))
-                (kill-buffer (find-buffer-visiting diary-file)))
+                (save-excursion
+                  (set-buffer (find-buffer-visiting diary-file))
+                  (subst-char-in-region (point-min) (point-max) ?\^M ?\n t)
+                  (setq selective-display nil)
+                  (set-buffer-modified-p diary-modified)))
             (beep)
             (message "Can't read included diary file %s" diary-file)
             (sleep-for 2))
@@ -391,6 +403,10 @@ changing the variable `diary-include-string'."
          (msg (format "No diary entries for %s %s"
                       (concat date-string (if holiday-list ":" ""))
                       (mapconcat 'identity holiday-list "; "))))
+    (calendar-set-mode-line
+     (concat "Diary for " date-string
+             (if holiday-list ": " "")
+             (mapconcat 'identity holiday-list "; ")))
     (if (or (not diary-entries-list)
             (and (not (cdr diary-entries-list))
                  (string-equal (car (cdr (car diary-entries-list))) "")))
@@ -406,10 +422,6 @@ changing the variable `diary-include-string'."
           (setq buffer-read-only t)
           (display-buffer holiday-buffer)
           (message  "No diary entries for %s" date-string))
-      (calendar-set-mode-line
-       (concat "Diary for " date-string
-               (if holiday-list ": " "")
-               (mapconcat 'identity holiday-list "; ")))
       (display-buffer (find-buffer-visiting d-file))
       (message "Preparing diary...done"))))
 
@@ -490,10 +502,10 @@ This function is provided for optional use as the `diary-display-hook'."
                   (if date-holiday-list (insert ":  "))
                   (let* ((l (current-column))
                          (longest 0))
-                    (insert (mapconcat '(lambda (x)
-                                          (if (< longest (length x))
-                                              (setq longest (length x)))
-                                          x)
+                    (insert (mapconcat (lambda (x)
+					 (if (< longest (length x))
+					     (setq longest (length x)))
+					 x)
                                        date-holiday-list
                                        (concat "\n" (make-string l ? ))))
                     (insert ?\n (make-string (+ l longest) ?=) ?\n)))))
@@ -585,7 +597,7 @@ is created."
 
 
 
-(defcustom diary-mail-addr 
+(defcustom diary-mail-addr
   (if (boundp 'user-mail-address) user-mail-address nil)
   "*Email address that `diary-mail-entries' will send email to."
   :group 'diary
@@ -614,39 +626,30 @@ emacs -batch \\
 -eval \"(setq diary-mail-days 3 \\
              european-calendar-style t \\
              diary-mail-addr \\\"user@host.name\\\" )\" \\
--l diary-lib -f diary-mail-entries 
+-l diary-lib -f diary-mail-entries
 at -f diary-rem.sh 0200 tomorrow
 
 You may have to tweak the syntax of the `at' command to suit your
 system.  Alternatively, you can specify a cron entry:
 0 1 * * * diary-rem.sh
 to run it every morning at 1am."
-  (interactive "p")
-  (let ((text nil)
-	;; Use the fancy-diary-display as it doesn't hide rest of
-	;; diary file with ^M characters.  It also looks nicer.
-	(diary-display-hook 'fancy-diary-display))	
-    (if (not current-prefix-arg)
-	(setq ndays diary-mail-days))
-    (calendar)
-    (view-diary-entries ndays)
-    (set-buffer fancy-diary-buffer)
-    (setq text (buffer-substring (point-min) (point-max)))
-
-    ;; Now send text as a mail message.
+  (interactive "P")
+  (let* ((diary-display-hook 'fancy-diary-display)
+         (diary-list-include-blanks t)
+         (text (progn (list-diary-entries (calendar-current-date)
+                                          (if ndays ndays diary-mail-days))
+                      (set-buffer fancy-diary-buffer)
+                      (buffer-substring (point-min) (point-max)))))
     (mail)
-    (mail-to)
-    (insert diary-mail-addr)
-    (mail-subject)
-    (insert "Diary entries generated ")
-    (insert (format-time-string "%a %d %b %Y" (current-time)))
-    (mail-text)
-    (insert text)
-    (mail-send-and-exit nil)
-    (exit-calendar)))
+    (mail-to) (insert diary-mail-addr)
+    (mail-subject) (insert "Diary entries generated "
+                           (calendar-date-string (calendar-current-date)))
+    (mail-text) (insert text)
+    (mail-send-and-exit nil)))
+
 
 (defun diary-name-pattern (string-array &optional fullname)
-  "Convert an STRING-ARRAY, an array of strings to a pattern.
+  "Convert a STRING-ARRAY, an array of strings to a pattern.
 The pattern will match any of the strings, either entirely or abbreviated
 to three characters.  An abbreviated form will match with or without a period;
 If the optional FULLNAME is t, abbreviations will not match, just the full
@@ -770,7 +773,7 @@ After the entries are marked, the hooks `nongregorian-diary-marking-hook' and
                                    (calendar-make-alist
                                     calendar-day-name-array
                                     0
-                                    '(lambda (x) (substring x 0 3))))))
+                                    (lambda (x) (substring x 0 3))))))
                           (if mm-name
                               (if (string-equal mm-name "*")
                                   (setq mm 0)
@@ -780,7 +783,7 @@ After the entries are marked, the hooks `nongregorian-diary-marking-hook' and
                                             (calendar-make-alist
                                              calendar-month-name-array
                                              1
-                                             '(lambda (x) (substring x 0 3)))
+                                             (lambda (x) (substring x 0 3)))
                                             )))))
                           (mark-calendar-date-pattern mm dd yy))))
                     (setq d (cdr d))))
@@ -956,7 +959,7 @@ after those with times."
   :type 'integer
   :group 'diary
   :version "20.3")
- 
+
 (defun diary-entry-time (s)
   "Time at the beginning of the string S in a military-style integer.  For
 example, returns 1325 for 1:25pm.  Returns `diary-unknown-time' (default value
@@ -964,20 +967,20 @@ example, returns 1325 for 1:25pm.  Returns `diary-unknown-time' (default value
 XX:XX (military time), and XXam, XXAM, XXpm, XXPM, XX:XXam, XX:XXAM XX:XXpm,
 or XX:XXPM."
   (let ((case-fold-search nil))
-    (cond ((string-match;; Military time  
-	    "^[ \t]*\\([0-9]?[0-9]\\):?\\([0-9][0-9]\\)\\(\\>\\|[^ap]\\)" s)
+    (cond ((string-match;; Military time
+	    "\\`[ \t\n\\^M]*\\([0-9]?[0-9]\\):?\\([0-9][0-9]\\)\\(\\>\\|[^ap]\\)" s)
 	   (+ (* 100 (string-to-int
 		      (substring s (match-beginning 1) (match-end 1))))
 	      (string-to-int (substring s (match-beginning 2) (match-end 2)))))
 	  ((string-match;; Hour only  XXam or XXpm
-	    "^[ \t]*\\([0-9]?[0-9]\\)\\([ap]\\)m\\>" s)
+	    "\\`[ \t\n\\^M]*\\([0-9]?[0-9]\\)\\([ap]\\)m\\>" s)
 	   (+ (* 100 (% (string-to-int
 			   (substring s (match-beginning 1) (match-end 1)))
 			  12))
 	      (if (equal ?a (downcase (aref s (match-beginning 2))))
 		  0 1200)))
 	  ((string-match;; Hour and minute  XX:XXam or XX:XXpm
-	    "^[ \t]*\\([0-9]?[0-9]\\):\\([0-9][0-9]\\)\\([ap]\\)m\\>" s)
+	    "\\`[ \t\n\\^M]*\\([0-9]?[0-9]\\):\\([0-9][0-9]\\)\\([ap]\\)m\\>" s)
 	   (+ (* 100 (% (string-to-int
 			   (substring s (match-beginning 1) (match-end 1)))
 			  12))
@@ -986,6 +989,8 @@ or XX:XXPM."
 		  0 1200)))
 	  (t diary-unknown-time))));; Unrecognizable
 
+;; Unrecognizable
+
 (defun list-sexp-diary-entries (date)
   "Add sexp entries for DATE from the diary file to `diary-entries-list'.
 Also, Make them visible in the diary file.  Returns t if any entries were
@@ -1121,7 +1126,7 @@ A number of built-in functions are available for this type of diary entry:
                   will appear on the proper Hebrew-date anniversary and on the
                   day before.  (If `european-calendar-style' is t, the order
                   of the parameters should be changed to DAY, MONTH, YEAR.)
-                  
+
       %%(diary-rosh-hodesh)
                   Diary entries will be made on the dates of Rosh Hodesh on
                   the Hebrew calendar.  Note that since there is no text, it
@@ -1281,9 +1286,10 @@ An optional parameter DAY means the Nth DAYNAME on or after/before MONTH DAY."
               (d2 (extract-calendar-day last))
               (y2 (extract-calendar-year last)))
 	 (if (or (and (= m1 m2)	; only possible base dates in one month
-		      (or (and (listp month) (memq m1 month))
-			  (eq month t)
-			  (= m1 month))
+		      (or (eq month t)
+			  (if (listp month)
+                              (memq m1 month)
+			    (= m1 month)))
 		      (let ((d (or day (if (> n 0)
 					   1
 					 (calendar-last-day-of-month m1 y1)))))
@@ -1294,22 +1300,25 @@ An optional parameter DAY means the Nth DAYNAME on or after/before MONTH DAY."
 		      (or
 		       ;; m1, d1 works as a base date
 		       (and
-			(or (and (listp month) (memq m1 month))
-			    (eq month t)
-			    (= m1 month))
+			(or (eq month t)
+			    (if (listp month)
+                                (memq m1 month)
+			      (= m1 month)))
 			(<= d1 (or day (if (> n 0)
 					   1
 					 (calendar-last-day-of-month m1 y1)))))
 		       ;; m2, d2 works as a base date
-		       (and (or (and (listp month) (memq m2 month))
-				(eq month t)
-				(= m2 month))
+		       (and (or (eq month t)
+				(if (listp month)
+                                    (memq m2 month)
+				  (= m2 month)))
 			    (<= (or day (if (> n 0)
 					    1
 					  (calendar-last-day-of-month m2 y2)))
 				d2)))))
 	     entry))))
 
+
 (defun diary-anniversary (month day year)
   "Anniversary diary entry.
 Entry applies if date is the anniversary of MONTH, DAY, YEAR if
@@ -1337,8 +1346,9 @@ to be March 1 in non-leap years."
   "Cycle diary entry--entry applies every N days starting at MONTH, DAY, YEAR.
 If `european-calendar-style' is t, parameters are N, DAY, MONTH, YEAR.
 ENTRY can contain `%d' or `%d%s'; the %d will be replaced by the number of
-years since the MONTH DAY, YEAR and the %s will be replaced by the ordinal
-ending of that number (that is, `st', `nd', `rd' or `th', as appropriate."
+repetitions since the MONTH DAY, YEAR and %s will be replaced by the
+ordinal ending of that number (that is, `st', `nd', `rd' or `th', as
+appropriate."
   (let* ((d (if european-calendar-style
                 month
               day))
@@ -1372,7 +1382,7 @@ ending of that number (that is, `st', `nd', `rd' or `th', as appropriate."
     diary-entry)
   "*Pseudo-pattern giving form of reminder messages in the fancy diary
 display.
- 
+
 Used by the function `diary-remind', a pseudo-pattern is a list of
 expressions that can involve the keywords `days' (a number), `date' (a list of
 month, day, year), and `diary-entry' (a string)."
@@ -1418,7 +1428,7 @@ marked on the calendar."
   "Add the entry (DATE STRING SPECIFIER) to `diary-entries-list'.
 Do nothing if DATE or STRING is nil."
   (and date string
-       (setq diary-entries-list 
+       (setq diary-entries-list
              (append diary-entries-list (list (list date string specifier))))))
 
 (defun make-diary-entry (string &optional nonmarking file)
@@ -1426,7 +1436,15 @@ Do nothing if DATE or STRING is nil."
 If omitted, NONMARKING defaults to nil and FILE defaults to diary-file."
   (find-file-other-window
    (substitute-in-file-name (if file file diary-file)))
+  (widen)
   (goto-char (point-max))
+  (when (let ((case-fold-search t))
+          (search-backward "Local Variables:"
+                           (max (- (point-max) 3000) (point-min))
+                           t))
+    (beginning-of-line)
+    (insert "\n")
+    (previous-line 1))
   (insert
    (if (bolp) "" "\n")
    (if nonmarking diary-nonmarking-symbol "")
@@ -1520,7 +1538,7 @@ Prefix arg will make the entry nonmarking."
      (format "%s(diary-cyclic %d %s)"
              sexp-diary-entry-symbol
              (calendar-read "Repeat every how many days: "
-                            '(lambda (x) (> x 0)))
+                            (lambda (x) (> x 0)))
              (calendar-date-string (calendar-cursor-to-date t) nil t))
      arg)))
 
