@@ -1,9 +1,8 @@
 ;;; mh-utils.el --- mh-e code needed for both sending and reading
-;; Time-stamp: <95/10/22 17:58:16 gildea>
 
-;; Copyright (C) 1993, 1995, 1997 Free Software Foundation, Inc.
+;; Copyright (C) 1993, 1995, 1997, 2000, 2001 Free Software Foundation, Inc.
 
-;; This file is part of mh-e, part of GNU Emacs.
+;; This file is part of GNU Emacs.
 
 ;; GNU Emacs is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -38,12 +37,24 @@
 (defvar mh-lib nil
   "Directory containing the MH library.
 This directory contains, among other things,
-the mhl program and the components file.")
+the components file.")
+
+(defvar mh-lib-progs nil
+  "Directory containing MH helper programs.
+This directory contains, among other things, 
+the mhl program.")
+
+(defvar mh-nmh-p nil
+  "Non-nil if nmh is installed on this system instead of MH")
 
 ;;;###autoload
 (put 'mh-progs 'risky-local-variable t)
 ;;;###autoload
 (put 'mh-lib 'risky-local-variable t)
+;;;###autoload
+(put 'mh-lib-progs 'risky-local-variable t)
+;;;###autoload
+(put 'mh-nmh-p 'risky-local-variable t)
 
 ;;; User preferences:
 
@@ -205,16 +216,16 @@ also show it in a separate Show window."
   ;; If SAVE-MODIFICATION-FLAG-P is non-nil, the buffer's modification
   ;; flag is unchanged, otherwise it is cleared.
   (setq save-modification-flag-p (car save-modification-flag-p)) ; CL style
-  (` (prog1
-	 (let ((mh-folder-updating-mod-flag (buffer-modified-p))
-	       (buffer-read-only nil)
-	       (buffer-file-name nil))	;don't let the buffer get locked
-	   (prog1
-	       (progn
-		 (,@ body))
-	     (mh-set-folder-modified-p mh-folder-updating-mod-flag)))
-       (,@ (if (not save-modification-flag-p)
-	       '((mh-set-folder-modified-p nil)))))))
+  `(prog1
+       (let ((mh-folder-updating-mod-flag (buffer-modified-p))
+	     (buffer-read-only nil)
+	     (buffer-file-name nil))	;don't let the buffer get locked
+	 (prog1
+	     (progn
+	       ,@body)
+	   (mh-set-folder-modified-p mh-folder-updating-mod-flag)))
+     ,@(if (not save-modification-flag-p)
+	   '((mh-set-folder-modified-p nil)))))
 
 (put 'with-mh-folder-updating 'lisp-indent-hook 1)
 
@@ -223,13 +234,13 @@ also show it in a separate Show window."
   ;; Display buffer SHOW-BUFFER in other window and execute BODY in it.
   ;; Stronger than save-excursion, weaker than save-window-excursion.
   (setq show-buffer (car show-buffer))	; CL style
-  (` (let ((mh-in-show-buffer-saved-window (selected-window)))
-       (switch-to-buffer-other-window (, show-buffer))
-       (if mh-bury-show-buffer (bury-buffer (current-buffer)))
-       (unwind-protect
-	   (progn
-	     (,@ body))
-	 (select-window mh-in-show-buffer-saved-window)))))
+  `(let ((mh-in-show-buffer-saved-window (selected-window)))
+     (switch-to-buffer-other-window ,show-buffer)
+     (if mh-bury-show-buffer (bury-buffer (current-buffer)))
+     (unwind-protect
+	 (progn
+           ,@body)
+       (select-window mh-in-show-buffer-saved-window))))
 
 (put 'mh-in-show-buffer 'lisp-indent-hook 1)
 
@@ -436,7 +447,7 @@ Type \"\\[mh-header-display]\" to see the message with all its headers."
   ;; window.  (Commands like save-some-buffers can make this false.)
   (if (eq (get-buffer-window (current-buffer))
 	  (selected-window))
-      (recenter (if arg arg '(t)))))
+      (recenter (if arg arg '(4)))))
 
 
 (defun mh-delete-line (lines)
@@ -529,14 +540,22 @@ Non-nil third argument means not to show the message."
 	     (end-of-line)
 	     (buffer-substring start (point)))))))
 
-(defvar mail-user-agent 'mh-e-user-agent) ;from reporter.el 3.2
+(defvar mail-user-agent)
+(defvar read-mail-command)
+
+(defvar mh-find-path-run nil
+  "Non-nil if `mh-find-path' has been run already.")
 
 (defun mh-find-path ()
-  ;; Set mh-progs and mh-lib.
+  ;; Set mh-progs, mh-lib, and mh-libs-progs
   ;; (This step is necessary if MH was installed after this Emacs was dumped.)
   ;; From profile file, set mh-user-path, mh-draft-folder,
   ;; mh-unseen-seq, mh-previous-seq, mh-inbox.
   (mh-find-progs)
+  (unless mh-find-path-run
+    (setq mh-find-path-run t)
+    (setq read-mail-command 'mh-rmail)
+    (setq mail-user-agent 'mh-e-user-agent))
   (save-excursion
     ;; Be sure profile is fully expanded before switching buffers
     (let ((profile (expand-file-name (or (getenv "MH") "~/.mh_profile"))))
@@ -559,7 +578,7 @@ Non-nil third argument means not to show the message."
 	    (if (not (mh-folder-name-p mh-draft-folder))
 		(setq mh-draft-folder (format "+%s" mh-draft-folder)))
 	    (if (not (file-exists-p (mh-expand-file-name mh-draft-folder)))
-		(error "Draft folder \"%s\" not found.  Create it and try again."
+		(error "Draft folder \"%s\" not found.  Create it and try again"
 		       (mh-expand-file-name mh-draft-folder)))))
       (setq mh-inbox (mh-get-profile-field "Inbox:"))
       (cond ((not mh-inbox)
@@ -573,7 +592,6 @@ Non-nil third argument means not to show the message."
       (setq mh-previous-seq (mh-get-profile-field "Previous-Sequence:"))
       (if mh-previous-seq
 	  (setq mh-previous-seq (intern mh-previous-seq)))
-      (setq mail-user-agent 'mh-e-user-agent)
       (run-hooks 'mh-find-path-hook)))
   (and mh-auto-folder-collect
        (let ((mh-no-install t))		;only get folders if MH installed
@@ -587,7 +605,8 @@ Non-nil third argument means not to show the message."
 
 (defun mh-find-progs ()
   "Find the `inc' and `mhl' programs of MH.
-Set the `mh-progs' and `mh-lib' variables to the file names."
+Set the `mh-progs' and `mh-lib', and `mh-lib-progs' variables to the
+directory names."
   (or (and mh-progs (mh-file-command-p (expand-file-name "inc" mh-progs)))
       (setq mh-progs
 	    (or (mh-path-search exec-path "inc")
@@ -596,38 +615,51 @@ Set the `mh-progs' and `mh-lib' variables to the file names."
 				  "/usr/bin/mh/" ;Ultrix 4.2
 				  "/usr/new/mh/" ;Ultrix <4.2
 				  "/usr/contrib/mh/bin/" ;BSDI
+				  "/usr/pkg/bin/"	; NetBSD
 				  "/usr/local/bin/"
 				  )
 				"inc"))))
+  (or (null mh-progs)
+      (let ((mh-base mh-progs))
+	(while (let ((dir-name (file-name-nondirectory
+				(directory-file-name mh-base))))
+		 (or (string= "mh" dir-name)
+		     (string= "bin" dir-name)))
+	  (setq mh-base
+		(file-name-directory (directory-file-name mh-base))))
+	(or (and mh-lib 
+		 (file-exists-p (expand-file-name "components" mh-lib)))
+	    (setq mh-lib
+		  ;; Look for a lib directory roughly parallel to the bin
+		  ;; directory:  Strip any trailing `mh' or `bin' path
+		  ;; components, then look for lib/mh or mh/lib.
+		  (or (mh-path-search
+		       (mapcar (lambda (p) (expand-file-name p mh-base))
+			       '("lib/mh" "etc/nmh" "/etc/nmh" "mh/lib" "etc" "lib"))
+		       "components"
+		       'file-exists-p))))
+	(or (and mh-lib-progs
+		 (mh-file-command-p (expand-file-name "mhl" mh-lib-progs)))
+	    (setq mh-lib-progs
+		  (or (mh-path-search
+		       (mapcar (lambda (p) (expand-file-name p mh-base))
+			       '("lib/mh" "libexec/nmh" "lib/nmh" "mh/lib" "lib"))
+		       "mhl")
+		      (mh-path-search '("/usr/local/bin/mh/") "mhl")
+		      (mh-path-search exec-path "mhl") ;unlikely
+		      )))))
+  (unless (and mh-progs mh-lib mh-lib-progs)
+    (error "Cannot find the commands `inc' and `mhl' and the file `components'"))
+  (setq mh-nmh-p (not (null
+		       (or (string-match "nmh" mh-lib-progs)
+			   (string-match "nmh" mh-lib))))))
 
-  (or (and mh-lib (mh-file-command-p (expand-file-name "mhl" mh-lib)))
-      (null mh-progs)
-      (setq mh-lib
-	    ;; Look for a lib directory roughly parallel to the bin
-	    ;; directory:  Strip any trailing `mh' or `bin' path
-	    ;; components, then look for lib/mh or mh/lib.
-	    (or (let ((mh-base mh-progs))
-		  (while (let ((dir-name (file-name-nondirectory
-					  (directory-file-name mh-base))))
-			   (or (string= "mh" dir-name)
-			       (string= "bin" dir-name)))
-		    (setq mh-base
-			  (file-name-directory (directory-file-name mh-base))))
-		  (mh-path-search
-		   (list (expand-file-name "lib/mh/" mh-base)
-			 (expand-file-name "mh/lib/" mh-base))
-		   "mhl"))
-		(mh-path-search '("/usr/local/bin/mh/") "mhl")
-		(mh-path-search exec-path "mhl") ;unlikely
-		)))
-  (unless (and mh-progs mh-lib)
-    (error "Cannot find the commands `inc' and `mhl'")))
-
-(defun mh-path-search (path file)
+(defun mh-path-search (path file &optional func-p)
   ;; Search PATH, a list of directory names, for FILE.
   ;; Returns the element of PATH that contains FILE, or nil if not found.
   (while (and path
-	      (not (mh-file-command-p (expand-file-name file (car path)))))
+	      (not (funcall (or func-p 'mh-file-command-p)
+			    (expand-file-name file (car path)))))
     (setq path (cdr path)))
   (car path))
 
@@ -646,7 +678,7 @@ Set the `mh-progs' and `mh-lib' variables to the file names."
   ;; mh-exec-cmd will display to the user.
   ;; The MH 5 version of install-mh might try prompt the user
   ;; for information, which would fail here.
-  (mh-exec-cmd (expand-file-name "install-mh" mh-lib) "-auto")
+  (mh-exec-cmd (expand-file-name "install-mh" mh-lib-progs) "-auto")
   ;; now try again to read the profile file
   (erase-buffer)
   (condition-case err
@@ -692,6 +724,7 @@ Set the `mh-progs' and `mh-lib' variables to the file names."
   (setq mode-name mode-name-string)
   (force-mode-line-update t))
 
+(defvar mh-folder-hist nil)
 
 (defun mh-prompt-for-folder (prompt default can-create)
   ;; Prompt for a folder name with PROMPT.  Returns the folder's name as a
@@ -706,8 +739,8 @@ Set the `mh-progs' and `mh-lib' variables to the file names."
 	 read-name folder-name)
     (if (null mh-folder-list)
 	(mh-set-folder-list))
-    (while (and (setq read-name (completing-read prompt mh-folder-list
-					    nil nil "+"))
+    (while (and (setq read-name (completing-read prompt mh-folder-list nil nil
+						 "+" 'mh-folder-hist default))
 		(equal read-name "")
 		(equal default "")))
     (cond ((or (equal read-name "") (equal read-name "+"))
@@ -918,7 +951,7 @@ Set the `mh-progs' and `mh-lib' variables to the file names."
 (defun mh-exec-lib-cmd-output (command &rest args)
   ;; Execute MH library command COMMAND with ARGS.
   ;; Put the output into buffer after point.  Set mark after inserted text.
-  (apply 'mh-exec-cmd-output (expand-file-name command mh-lib) nil args))
+  (apply 'mh-exec-cmd-output (expand-file-name command mh-lib-progs) nil args))
 
 
 (defun mh-handle-process-error (command status)
@@ -944,7 +977,7 @@ Set the `mh-progs' and `mh-lib' variables to the file names."
 						 (end-of-line)
 						 (point))))
 	     (display-buffer (current-buffer))
-	     (error "%s failed with status %d.  See error message in other window."
+	     (error "%s failed with status %d.  See error message in other window"
 		    command status)))))))
 
 
